#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include <types.h>
#include <fhp_kernel.cu>


TYPE c1[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c2[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c3[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c4[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c5[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE c6[LATTICE_WIDTH * LATTICE_HEIGHT];

TYPE k1[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k2[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k3[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k4[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k5[LATTICE_WIDTH * LATTICE_HEIGHT];
TYPE k6[LATTICE_WIDTH * LATTICE_HEIGHT];

TYPE boundary[LATTICE_WIDTH * LATTICE_HEIGHT];

int mouse_x, mouse_y;
unsigned int next_rand;

int random_int()
{
    next_rand = next_rand * 1103515245 + 12345;
    return (unsigned int) (next_rand / 65536);
}

int flip_coin()
{
    int r = random_int();
    return r % 2;
}

void next()
{
    /*
     *   3      2     *    c      b
     *    \    /      *     \    /
     *     \  /       *      \  /
     * 4 ---[]--- 1   *  d ---[]--- a
     *     /  \       *      /  \
     *    /    \      *     /    \
     *   5      6     *    e      f
     *
     */
	
	TYPE *gpu_boundary;
	TYPE *gpu_c1, *gpu_c2, *gpu_c3, *gpu_c4, *gpu_c5, *gpu_c6;
	TYPE *gpu_k1, *gpu_k2, *gpu_k3, *gpu_k4, *gpu_k5, *gpu_k6;

	dim3 block(8, 8, 1);
	dim3 grid(LATTICE_WIDTH / block.x, LATTICE_HEIGHT / block.y, 1);

	cudaMalloc((void **) &gpu_c1, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_c2, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_c3, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_c4, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_c5, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_c6, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);

	cudaMalloc((void **) &gpu_k1, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_k2, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_k3, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_k4, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_k5, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);
	cudaMalloc((void **) &gpu_k6, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);

	cudaMalloc((void **) &gpu_boundary, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH);

	cudaMemcpy(gpu_c1, c1, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_c2, c2, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_c3, c3, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_c4, c4, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_c5, c5, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_c6, c6, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);

	cudaMemcpy(gpu_k1, k1, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_k2, k2, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_k3, k3, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_k4, k4, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_k5, k5, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);
	cudaMemcpy(gpu_k6, k6, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);

	cudaMemcpy(gpu_boundary, boundary, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyDeviceToHost);

	collision<<<grid, block>>>(gpu_c1, gpu_c2, gpu_c3, gpu_c4, gpu_c5, gpu_c6,
				   gpu_k1, gpu_k2, gpu_k3, gpu_k4, gpu_k5, gpu_k6,
				   gpu_boundary);	
	propagate<<<grid, block>>>(gpu_c1, gpu_c2, gpu_c3, gpu_c4, gpu_c5, gpu_c6,
				   gpu_k1, gpu_k2, gpu_k3, gpu_k4, gpu_k5, gpu_k6);
	
	cudaMemcpy(c1, gpu_c1, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(c2, gpu_c2, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(c3, gpu_c3, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(c4, gpu_c4, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(c5, gpu_c5, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(c6, gpu_c6, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);

	cudaMemcpy(k1, gpu_k1, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(k2, gpu_k2, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(k3, gpu_k3, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(k4, gpu_k4, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(k5, gpu_k5, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
	cudaMemcpy(k6, gpu_k6, LATTICE_WIDTH * LATTICE_HEIGHT * LENGTH, cudaMemcpyHostToDevice);
}

void draw_point(int x, int y, short value)
{
	if(value > 0) {
		glColor3f((float)value/(LENGTH*6.0), (float)value/(LENGTH * 6.0), 0.9);
	}
	else {
		glColor3f(0.0, 0.0, 0.0);
	}
    glVertex3f(x + WINDOW_WIDTH / 8, y + WINDOW_HEIGHT / 8, 0.0);
}

void display()
{
    int i, j;
    int x, y;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glBegin(GL_POINTS);
    for (x = 0; x < LATTICE_WIDTH; x++) {
        for (y = 0; y < LATTICE_HEIGHT; y++) {
            i = x * SPACING;
            j = y * SPACING;

            glColor3f(0.2, 0.0, 0.0);
            draw_point(i, j, c1[y * LATTICE_WIDTH + x] + c2[y * LATTICE_WIDTH + x] + c3[y * LATTICE_WIDTH + x] + 
                             c4[y * LATTICE_WIDTH + x] + c5[y * LATTICE_WIDTH + x] + c6[y * LATTICE_WIDTH + x]);

            if (boundary[y* LATTICE_WIDTH + x] == 0) {
                glColor3f(1.0, 1.0, 1.0);
                glVertex3f(i     + WINDOW_WIDTH / 8, j + 1 + WINDOW_HEIGHT / 8, 0.0);
                glVertex3f(i - 1 + WINDOW_WIDTH / 8, j - 1 + WINDOW_HEIGHT / 8, 0.0);
                glVertex3f(i + 1 + WINDOW_WIDTH / 8, j - 1 + WINDOW_HEIGHT / 8, 0.0);
                glVertex3f(i     + WINDOW_WIDTH / 8, j - 1 + WINDOW_HEIGHT / 8, 0.0);
                glColor3f(0.0, 0.0, 0.0);
            }
        }
    }
    glEnd();

    glutSwapBuffers();
}

void idle()
{
    
	/*
	int i, j;
    for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = pow(2.0, LENGTH) - 1; //random_int( ) % 16;
                c2[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = pow(2.0, LENGTH) - 1; //random_int( ) % 16;
                c3[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = pow(2.0, LENGTH) - 1; //random_int( ) % 16;
                c4[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = pow(2.0, LENGTH) - 1; //random_int( ) % 16;
                c5[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = pow(2.0, LENGTH) - 1; //random_int( ) % 16;
                c6[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = pow(2.0, LENGTH) - 1; //random_int( ) % 16;
            }
        }
    }
	*/
	/*
	for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c2[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c3[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c4[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c5[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c6[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
            }
        }
    }
	*/
/*    printf("boundary at i=%d  is %d\n", WINDOW_WIDTH/3, boundary[WINDOW_WIDTH/3][0]);*/
    next();
    glutPostRedisplay();
}

void init()
{
    int i, j, k;
    next_rand = 0;

    for (i = 0; i < LATTICE_WIDTH; i++) {
        for (j = 0; j < LATTICE_HEIGHT; j++) {
            c1[j * LATTICE_WIDTH + i] = 0;
            c2[j * LATTICE_WIDTH + i] = 0;
            c3[j * LATTICE_WIDTH + i] = 0;
            c4[j * LATTICE_WIDTH + i] = 0;
            c5[j * LATTICE_WIDTH + i] = 0;
            c6[j * LATTICE_WIDTH + i] = 0;
            /*
			if(random_int( ) % 999 == 0) {
                c1[i][j] = random_int( ) % (int) pow(2.0, LENGTH);
                c2[i][j] = random_int( ) % (int) pow(2.0, LENGTH);
                c3[i][j] = random_int( ) % (int) pow(2.0, LENGTH);
                c4[i][j] = random_int( ) % (int) pow(2.0, LENGTH);
                c5[i][j] = random_int( ) % (int) pow(2.0, LENGTH);
                c6[i][j] = random_int( ) % (int) pow(2.0, LENGTH);
            }
			*/
            boundary[j * LATTICE_WIDTH + i] = 0xF;
        }
    }

    for (i = 0; i < LATTICE_WIDTH; i++) {
        boundary[(LATTICE_HEIGHT - 1) * LATTICE_WIDTH + i] = 0;
        boundary[i] = 0;
    }

    for (j = 0; j < LATTICE_HEIGHT; j++) {
        boundary[j * LATTICE_WIDTH] = 0;
        boundary[j * LATTICE_WIDTH - 1] = 0;
    }

	
	for(k = 0; k < 10; k++) {
		for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
			for (j = 0; j < LATTICE_WIDTH / 20; j++) {
				if (1) {
					c1[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c2[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c3[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c4[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c5[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
					c6[k * j * LATTICE_WIDTH + k * i] = random_int( ) % 16;
				}
			}
		}
	}
	
    for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c2[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c3[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c4[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c5[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
                c6[j * LATTICE_WIDTH/2 + i + LATTICE_WIDTH / 2] = random_int( ) % 16;
            }
        }
    }
	/*
	for (i = 0; i < LATTICE_HEIGHT / 20; i++) {
        for (j = 0; j < LATTICE_WIDTH / 20; j++) {
            if (1) {
                c1[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c2[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c3[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c4[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c5[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
                c6[i + LATTICE_WIDTH / 2][j + LATTICE_HEIGHT / 2] = random_int( ) % (int) pow(2.0, LENGTH);
            }
        }
    }
	*/
    glClearColor(0.0, 0.0, 0.0, 0.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, WINDOW_WIDTH, 0.0, WINDOW_HEIGHT, -20.0, 20.0);

    glMatrixMode(GL_MODELVIEW);

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glShadeModel(GL_SMOOTH);
    glPointSize(POINT_SIZE);
}

void mouse(int button, int state, int x, int y)
{
    mouse_x = x;
    mouse_y = y;
}

int main(int argc, char **argv)
{
    mouse_x = 0;
    mouse_y = 0;

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
    glutCreateWindow("fhp");
    glEnable(GL_DEPTH_TEST);
    init();
    glutDisplayFunc(display);
    glutIdleFunc(idle);
    glutMainLoop();
    return 0;
}
